% In this script we define a state space system modeling cars i and i-1.
% Then this model will be used in car i to detect and estimate the MITM
% attack

%% Simulation settings
runtime = 60; %[s]
stepsize = 1/1000; %[s]

plot_param = 0;

%% Simulation Scenario (Input & Attack)
input_time = [1 6];
input_mag = 1;

attack_time = 2;
attack_dur = 10;%58;
if ~AttackSweep
attack_mag = 2.8;
end
%x=[p_l v_l a_l p_f v_f a_f] 
x_0 = [-40;15;0;-50;15;0];%[m m/s m/s^2 m m/s m/s^2]

%% General 'True' model of both cars
%_f is used for the follower car, _l for the leader car.
%tau is engine time constant
tau_l = 0.11;%[s]
tau_f = 0.1;%[s]
L_f = 4;%[m]
L_l = 4;%[m] Only used to determine whether a crash occured (in intersection control)

%Model: x_dot = Ax+Bu+D(du); y=Cx+Fzeta
A_l = [[0 1 0];[0 0 1];[0 0 -1/tau_l]]; %leader
A_f = [[0 1 0];[0 0 1];[0 0 -1/tau_f]]; %follower
A = [[A_l zeros(3)];[zeros(3) A_f]]; %combined

B_l = [0;0;1/tau_l]; %leader
B_f = [0;0;1/tau_f]; %follower
B = [[B_l zeros(3,1)];[zeros(3,1) B_f]]; %combined

D = [-B_l;zeros(3,1)];

C = [[1 0 0 -1 0 0];[0 1 0 0 -1 0];[0 0 0 0 1 0];[0 0 0 0 0 1]]; %measurement of follower car (distance,rel. velocity, fol. velocity, fol. acceleration)

F=eye(4); %all sensors are affected by noise

%sensors of follower car
if ~UncSweep
    sensor.var = [0.05;0.1;0.01;0.05].^2; %variance[0.36; 0.36; 0.36; 0.36].^2;%
end
sensor.T = [0.04;0.04;0.01;0.01]; %sample time
%% Controller in follower car
kp = 0.2; %Proportional gain in CACC control algorithm
kd = 0.7; %Differential gain in CACC control algorithm
h = 0.7; %Lead time defined for reference distance
r = 1.5; %dinstance at standstil for reference distance
%d_r,i = r_i+h_i*v_i

%Controller:z_dot=A_cz+B_cy+D_cu+const; u[2] = C_cz; u[1]=recieved
A_c = -1/h;
B_c = [kp/h kd/h -kp -kd];
D_c = [1/h 0];
const = kp/h*r;

C_c = 1;
%% Observer in Follower car
if ~(UncSweep || AttackSweep)
    %observer design parameters
    M=diag([2.3500 3.300 2.160 3.6000]);%diag([0.5,11.5,0.2,2]);%+10*eye(4);%diag([2,12,1,5]);%Diagonal matrix 
    K=0.1*eye(4);%2%filtergain
end
A_22s=-eye(4)*0.1;%stable replacement A_22

%uncertainty bounds
zeta2_max = 3*sqrt(sensor.var);%maximum noise on sensors
maxacc_l = 5; %maximum input of leader car

%the model the observer uses
obs.hattau_l = 0.1;
obs.tau_f = 0.1;

zeta1_max = abs(tau_l/obs.hattau_l-1)*2*maxacc_l;

%Model: hatx_dot = Ahatx+Bu; haty=Chatx
obs.A_l = [[0 1 0];[0 0 1];[0 0 -1/obs.hattau_l]];
obs.A_f = [[0 1 0];[0 0 1];[0 0 -1/obs.tau_f]];
obs.A = [[obs.A_l zeros(3)];[zeros(3) obs.A_f]];

obs.B_l = [0;0;1/obs.hattau_l];
obs.B_f = [0;0;1/obs.tau_f];
obs.B = [[obs.B_l zeros(3,1)];[zeros(3,1) obs.B_f]];

obs.C = [[1 0 0 -1 0 0];[0 1 0 0 -1 0];[0 0 0 0 1 0];[0 0 0 0 0 1]];

%matrices needed to estimate attack
obs.D = [-obs.B_l;zeros(3,1)];

%matrices needed to bound uncertainty in estimation
obs.E = [zeros(2,1);1/obs.hattau_l;zeros(3,1)];

obs.F=eye(4);

%model transformation
T_1 = [[1 0 0 0 0 0];[0 0 1 0 0 0]];
T_C = [T_1;obs.C];

T_obs = [eye(2)];
T_B = [[T_obs zeros(2,4)];[zeros(4,2) eye(4)]]; %is already in this form, one unobservable state.

T_A = eye(6); %is already in this form, A_11 is stable.

T=T_A*T_B*T_C;

obs.C = obs.C*T^-1;
obs.D = T*obs.D;
obs.B = T*obs.B;
obs.A = T*obs.A*T^-1;
obs.E = T*obs.E;

temp = T*x_0;
Thatx_0 = [zeros(2,1);temp(3:end)];

%partition
p = rank(obs.C);%#measurements
n = size(obs.A,1);%#states

obs.A_11 = obs.A(1:n-p,1:n-p);
obs.A_21 = obs.A(n-p+1:end,1:n-p);
obs.A_22 = obs.A(n-p+1:end,n-p+1:end);
obs.A_12 = obs.A(1:n-p,n-p+1:end);

obs.B_1 = obs.B(1:n-p,:);
obs.B_2 = obs.B(n-p+1:end,:);

obs.D_1 = obs.D(1:n-p);
obs.D_2 = obs.D(n-p+1:end);

obs.E_1 = obs.E(1:n-p);
obs.E_2 = obs.E(n-p+1:end);

U_max2 = vecnorm(A_22s*obs.F,2,2)*norm(zeta2_max)+...
        vecnorm((obs.A_22-A_22s)*obs.F,2,2)*norm(zeta2_max)+...
        vecnorm(obs.E_2,2,2)*norm(zeta1_max)+...
        vecnorm(obs.A_21*pinv(obs.A_11)*obs.A_12*obs.F,2,2)*norm(zeta2_max)+...
        vecnorm(obs.A_21*pinv(obs.A_11)*obs.E_1,2,2)*norm(zeta1_max);%maximum uncertainty
U_max = abs(A_22s*obs.F)*abs(zeta2_max)+...
    abs((obs.A_22-A_22s)*obs.F)*abs(zeta2_max)+...
    abs(obs.E_2)*abs(zeta1_max)+...
    abs(obs.A_21*pinv(obs.A_11)*obs.A_12*obs.F)*abs(zeta2_max)+...
    abs(obs.A_21*pinv(obs.A_11)*obs.E_1)*abs(zeta1_max);%maximum uncertainty
e2dot_min = diag(M)-U_max;
e2dot_max = diag(M)+U_max;
r_e = e2dot_max./e2dot_min;
t_max = 2*F*zeta2_max./e2dot_min;
%% Detection time (no noise)
A_e = [[obs.A_11 zeros(2,4)];[-K*obs.A_21 -K]];
E_e = [-obs.E_1;K*obs.E_2];
D_e = [-obs.D_1;K*obs.D_2];

e_10=pinv(obs.A_11)*abs(obs.A_12*obs.F)*zeta2_max+...
    pinv(obs.A_11)*abs(obs.E_1)*zeta1_max;